/*****************************************************************************
 * Copyright (c) 2016 Christian W. Damus and others.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  Christian W. Damus - Initial API and implementation
 *  
 *****************************************************************************/
package org.eclipse.papyrusrt.umlrt.core.utils;

import static org.eclipse.papyrusrt.umlrt.core.types.ElementTypeUtils.getConfigureAsCommand;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.gmf.runtime.common.core.command.ICommand;
import org.eclipse.papyrusrt.umlrt.core.types.ElementTypeUtils;
import org.eclipse.papyrusrt.umlrt.core.types.IUMLRTElementTypes;
import org.eclipse.papyrusrt.umlrt.core.types.UMLRTElementTypesEnumerator;
import org.eclipse.papyrusrt.umlrt.profile.statemachine.UMLRealTimeStateMach.RTRegion;
import org.eclipse.papyrusrt.umlrt.profile.statemachine.UMLRealTimeStateMach.UMLRealTimeStateMachPackage;
import org.eclipse.uml2.uml.Pseudostate;
import org.eclipse.uml2.uml.PseudostateKind;
import org.eclipse.uml2.uml.Region;
import org.eclipse.uml2.uml.State;
import org.eclipse.uml2.uml.Vertex;

/**
 * Utilities for inspecting and working with regions in an UML-RT model.
 */
public class RegionUtils {

	/**
	 * Queries whether a {@code region} is stereotyped as an UML-RT region.
	 * 
	 * @param region
	 *            a region
	 * 
	 * @return whether it is an <tt>RTRegion</tt>
	 */
	public static boolean isRTRegion(Region region) {
		return ElementTypeUtils.matches(region, IUMLRTElementTypes.RT_REGION_ID);
	}

	/**
	 * Obtains a command that ensures that a {@code region} is stereotyped as an
	 * UML-RT region.
	 * 
	 * @param region
	 *            a region
	 * 
	 * @return the command
	 */
	public static ICommand getConfigureAsRTRegionCommand(Region region) {
		return getConfigureAsCommand(UMLRTElementTypesEnumerator.RT_REGION, region);
	}

	/**
	 * Applies the {@link RTRegion} stereotype to a {@code region}, if it is
	 * not already applied.
	 * 
	 * @param region
	 *            a region
	 * @return the region with the stereotype applied
	 */
	public static Region applyStereotype(Region region) {
		return UMLRTProfileUtils.ensureStereotype(region,
				UMLRealTimeStateMachPackage.Literals.RT_REGION);
	}

	public static List<Pseudostate> getInitialPseudoStates(Region region) {
		List<Pseudostate> pseudoStates = new ArrayList<>();
		for (Vertex vertex : region.getSubvertices()) {
			if (isInitialPseudoState(vertex)) {
				pseudoStates.add((Pseudostate) vertex);
			}
		}
		return pseudoStates;
	}

	public static List<State> getStates(Region region) {
		List<State> states = new ArrayList<>();
		for (Vertex vertex : region.getSubvertices()) {
			if (vertex instanceof State) {
				states.add((State) vertex);
			}
		}
		return states;
	}

	public static boolean isInitialPseudoState(Vertex vertex) {
		if (vertex instanceof Pseudostate) {
			Pseudostate ps = (Pseudostate) vertex;
			if (ps.getKind() == PseudostateKind.INITIAL_LITERAL) {
				return true;
			}
		}
		return false;
	}
}
